con
'un comment for 96x96 pixels
  lcd_p = 96              'pixels per line
  lcd_l = 96              'lines in a display }
        

'un comment for 128x128 pixels
  {lcd_p = 128              'pixels per line
  lcd_l = 128              'lines in a display }
  
'calculated constants
  lcd_longs = lcd_p*lcd_l/32 'longs to alocate for buffer.  Assumes lcd_p is divisable by 32
  line_longs = lcd_p/32         '#number of longs in a line.  Assumes lcd_p is divisable by 32
  'msb_mask = |< 31               'just the MSB set}

var
  long  cog


pub start(lcd_pp, lcd_ll)
''initialises and starts up the frame-buffer graphics cog

  if lcd_pp <> lcd_p
    abort
  if lcd_ll <> lcd_l
    abort
  'lcd_p := lcd_pp
  'lcd_l := lcd_ll
  'lcd_longs := lcd_p*lcd_l/32
  'line_longs := lcd_p/32

  cog := cognew(@entry, @command) + 1           'start the new cog now, assembly cog at "entry" label.


pub stop
  if cog
    cogstop(cog~ - 1)                                   'if the driver is already running, stop the cog  

pub wait_cmd
  ''wait for the most recent command to complete.
  ''only waits if the GFX cog is loaded.
  if cog                        'only wait if cog is loaded
    repeat while command         'wait for completion
  result := row_col

pub life_generation_nb(life_array_ptr)
''calculate the next generation of the life array
''non-blocking
  'longmove(@life_BC_end,@life_array,line_longs)
  'longmove(@life_BC_start,@life_array + 4*(lcd_longs - line_longs),line_longs)
  data_ptr := life_array_ptr
  row_col := 0
  command := life_cmd

pub life_generation(ptr)
''calculate the next generation of the life array
''blocking
  life_generation_nb(ptr)
  result := wait_cmd

pub census_nb(ptr)
  data_ptr := ptr
  row_col := 0
  command := census_cmd

pub census(ptr)
  census_nb(ptr)
  result := wait_cmd



dat
org           0
entry   'start of graphics code.

cmd_return    mov       command, #0            'clear command            
              wrlong    command, par           'clear command in hub as well
              mov       t1,     par             'setup temporary hub read pointers
              sub       t1,     #4              'one long before par => row_col
              mov       t2,     par
              sub       t2,     #8              'two longs before par => data_ptr
              mov       t3,     #153
              add       t3,     cnt  
cmd_wait       'wait for command <> 0
              waitcnt   t3,     #153            '(optional) use waitcnt to slow down the command scanning loop and save power.
              rdlong    command,     par             'read in command
              tjz       command,     #cmd_wait        'if command is zero, read it again.
              'read the command parameters.
              rdlong    row_col, t1             'read in multiplexed parameter variable.
              rdlong    data_ptr, t2            'read in the source data HUB pointer.  
              'branch to the correct code.
              cmps      command,     life_cmd        wz              
        if_z  call       #life_asm                       'calculate the next life generation
              cmps      command,     census_cmd     wz
        if_z  call       #census_asm                    'count the number of live cells in the current generation
              cmps      command,     text_cmd        wz
        if_z  call      #text_asm                       'write a letter or number to the frame-buffer
              jmp       #cmd_return             'clear the command and wait for a new command.

''calculates the next life generation using the frame buffer @data_ptr as the last life generation
''needs one extra row of pixel buffer before and after the frame buffer
''has no effect on row_col
''
''strip to basics and get the row selection loop working with an XOR bassed cellular automata.

life_asm

{{test line move/rotate calls}}
              'setup bottom boundary conditions
              mov       src_ptr, data_ptr
              mov       dest_ptr, data_ptr
              add       dest_ptr, lcd_bytes_asm
              call      #mov_line_hub                
              'setup top boundary condition
              mov       dest_ptr, data_ptr
              sub       dest_ptr, #(line_longs*4)
              mov       src_ptr,  data_ptr
              add       src_ptr,  lcd_bytes_asm
              sub       src_ptr,  #(line_longs*4)
              call      #mov_line_hub              

              'initial fill of direction buffers
              'row-1 
              mov       src_ptr, data_ptr
              sub       src_ptr, #(line_longs*4)
              mov       dest_ptr, #n
              call      #mov_line_from_hub
              mov       src_ptr, #n
              mov       dest_ptr, #nw
              call      #mov_line
              call      #ror_line
              mov       src_ptr, #n
              mov       dest_ptr, #ne
              call      #mov_line
              call      #rol_line
              'row
              mov       src_ptr, data_ptr
              mov       dest_ptr, #ct
              call      #mov_line_from_hub
              mov       src_ptr, #ct
              mov       dest_ptr, #w
              call      #mov_line
              call      #ror_line
              mov       src_ptr, #ct
              mov       dest_ptr, #e
              call      #mov_line
              call      #rol_line
              'row + 1
              mov       src_ptr, data_ptr
              add       src_ptr, #(line_longs*4)
              mov       dest_ptr, #s
              call      #mov_line_from_hub
              mov       src_ptr, #s
              mov       dest_ptr, #sw
              call      #mov_line
              call      #ror_line
              mov       src_ptr, #s
              mov       dest_ptr, #se
              call      #mov_line
              call      #rol_line
              
              'start of solver loops
              mov       row_ptr, data_ptr
              mov       row,    #0
              mov       row_idx, #lcd_l
:row_loop     'start of row loop  (get this working first!)
               {START collumn code and cellular automata}
              mov       col_idx, #line_longs    'loop several times
              mov       col,    #0
:col_loop
              mov       col_ptr, row_ptr
              mov       t1,     col
              shl       t1,     #2              'multiply by 4
              add       col_ptr, t1             'make the hub pointer                
              'ready to start add_loop for the 3-bit x 32 parallel adder.
              mov       s_3,    #0
              mov       s_2,    #0
              mov       add_ptr, #nw            'direction register starting cog address
              add       add_ptr, col            'offset to correct collumn    
              movs      :add_mod, add_ptr
              mov       add_idx, #7             'prep for 7 rounds of the adder loop
:add_mod      mov       s_1,    0-0             'preload s_1
:add_loop     add       add_ptr, #line_longs    'temp2 += line_longs (advance to next direction register)
              movs      :add_mod2, add_ptr     'temp := long[temp2]
              nop
:add_mod2     mov       t1,     0-0              
              mov       carry_1, t1             'carry_1 := temp & s_1
              and       carry_1, s_1
              xor       s_1,    t1              's_1 ^= temp
              mov       carry_2, carry_1        'carry_2 := carry_1 & s_2
              and       carry_2, s_2
              xor       s_2,    carry_1         's_2 ^= carry_1
              'mov       carry_3, carry_2        'carry_3 := carry_2 & s_3
              'and       carry_3, s_3            'enable to support full "life like" family of automata
              xor       s_3,    carry_2         's_3 ^= carry_2
              djnz      add_idx, #:add_loop
              'calculate and write back result.
              'long[col_ptr] := s_2 & !s_3 & (s_1 | long[@ct + col<<2] )
              add       add_ptr, #line_longs    'now points to "ct"
              movs      :add_mod3, add_ptr      'long[@ct + col<<2]
              nop
:add_mod3     mov       t1,     0-0
              or        t1,     s_1             '(s_1 | long[~~])
              andn      t1,     s_3             '& !s_3
              and       t1,     s_2             '& s_2
              wrlong    t1,     col_ptr         'long[col_ptr] := 
              
              'add loop done. advance to next collumn
              add       col,    #1
              djnz      col_idx, #:col_loop
               {END collumn code and cellular automata}
              'advance indexes to next row
              add       row, #1
              add       row_ptr, #(line_longs*4) 'shift hub pointer to next row
              'shift direction arrays north
              'nw := w
              mov       src_ptr, #w
              mov       dest_ptr, #nw
              call      #mov_line
              'n := ct
              mov       src_ptr, #ct
              mov       dest_ptr, #n
              call      #mov_line
              'ne := e
              mov       src_ptr, #e
              mov       dest_ptr, #ne
              call      #mov_line
              'w := sw
              mov       src_ptr, #sw
              mov       dest_ptr, #w
              call      #mov_line
              'ct := s
              mov       src_ptr, #s
              mov       dest_ptr, #ct
              call      #mov_line
              'e := se
              mov       src_ptr, #se
              mov       dest_ptr, #e
              call      #mov_line
              'grab new data from hub for southern most direction arrays
              's := life_array[row+1]
              mov       src_ptr, row_ptr
              add       src_ptr, #(line_longs*4)
              mov       dest_ptr, #s
              call      #mov_line_from_hub
              'sw := s -> 1
              mov       src_ptr, #s
              mov       dest_ptr, #sw
              call      #mov_line
              call      #ror_line
              'se := s <- 1
              mov       src_ptr, #s
              mov       dest_ptr, #se
              call      #mov_line
              call      #rol_line

              'loop
              djnz      row_idx, #:row_loop

life_asm_ret  ret

'pointers for the following composite instructions
src_ptr       long  0-0                         'base off of data_ptr for hub operations.
dest_ptr      long  0-0                         'use "mov dest_ptr, #COGVAR" for cog addresses
'composite instruction to move whole row/lines of pixel data around the hub
mov_line_hub
              mov       idx1,    #line_longs      'perpare to move a whole row
:loop         rdlong    t1,     src_ptr         'read in a long
              add       src_ptr, #4             'advance to next long
              wrlong    t1,     dest_ptr        'write out the long
              add       dest_ptr, #4            'advance to next long
              djnz      idx1,    #:loop          'loop for a whole row }             
mov_line_hub_ret  ret      

'composite instruction to copy whole row/lines of pixed data from the hub
mov_line_from_hub
              mov       idx1,    #line_longs      'perpare to move a whole row
              movd      :mod,   dest_ptr        'move value of dest_ptr to destination of the instruction at :mod
:loop         rdlong    t1,     src_ptr         'read in a long
              add       src_ptr, #4             'advance to next long
:mod          mov       0-0,    t1              'write t1 to COG memory
              add       :mod,   inc_d           'increment the destination
              djnz      idx1,    #:loop          'loop for a whole row
mov_line_from_hub_ret  ret

'composite instruction to copy whole row/lines of pixed data to the hub
mov_line_to_hub
              movs      :mod,   src_ptr        'move value of src_ptr to source of the instruction at :mod
              mov       idx1,    #line_longs      'perpare to move a whole row
:loop
:mod          mov       t1,    0-0              'read t1 from COG memory
              add       :mod,   inc_s           'increment the source
              wrlong    t1,     dest_ptr         'write out a long
              add       dest_ptr, #4             'advance to next long
              djnz      idx1,    #:loop          'loop for a whole row
mov_line_to_hub_ret  ret

'composite instructin to move a whole row/line of data around COG ram
mov_line
              movs      :mod,   src_ptr
              movd      :mod,   dest_ptr        'move value of dest_ptr to destination of the instruction at :mod
              mov       idx1,    #line_longs      'perpare to move a whole row
:loop         
:mod          mov       0-0,    0-0              'write move source long to dest lon in COG memory
              add       :mod,   inc_sd           'increment the source and destination
              djnz      idx1,    #:loop          'loop for a whole row
mov_line_ret  ret

'composite instruction to rotate a whole row/line one bit left
'only uses dest_ptr
ror_line
              movd      :mod1,  dest_ptr
              movd      :mod2,  dest_ptr
              mov       idx1,   #line_longs
:loop
:mod1         rcr       0-0,    #1              wc 'rotate through C by one
              add       :mod1,  inc_d           'increment address of long to rotate
              djnz      idx1,   #:loop          'loop for one row/line
:mod2         muxc      0-0,    msb_mask        'finish up and fix lsb of first long.
ror_line_ret  ret

'composite instruction to rotate a whole row/line one bit right
'only uses dest_ptr
rol_line
              mov       t1,     dest_ptr
              add       t1,     #(line_longs-1) 'advance to end of row/line
              movd      :mod1,  t1
              movd      :mod2,  t1
              mov       idx1,   #line_longs
:loop
:mod1         rcl       0-0,    #1              wc 'rotate through C by one
              sub       :mod1,  inc_d           'decrement address of long to rotate
              djnz      idx1,   #:loop          'loop for one row/line
:mod2         muxc      0-0,    #1              'finish up and fix lsb of first long.
rol_line_ret  ret

'composite instructin to XOR a whole row/line of data in COG ram
xor_line
              movs      :mod,   src_ptr
              movd      :mod,   dest_ptr        'move value of dest_ptr to destination of the instruction at :mod
              mov       idx1,    #line_longs      'perpare to move a whole row
:loop         
:mod          xor       0-0,    0-0              'XOR source long to dest lon in COG memory
              add       :mod,   inc_sd           'increment the source and destination
              djnz      idx1,    #:loop          'loop for a whole row
xor_line_ret  ret

'composite instructin to AND a whole row/line of data in COG ram
and_line
              movs      :mod,   src_ptr
              movd      :mod,   dest_ptr        'move value of dest_ptr to destination of the instruction at :mod
              mov       idx1,    #line_longs      'perpare to move a whole row
:loop         
:mod          and       0-0,    0-0              'XOR source long to dest lon in COG memory
              add       :mod,   inc_sd           'increment the source and destination
              djnz      idx1,    #:loop          'loop for a whole row
and_line_ret  ret

'composite instructin to OR a whole row/line of data in COG ram
or_line
              movs      :mod,   src_ptr
              movd      :mod,   dest_ptr        'move value of dest_ptr to destination of the instruction at :mod
              mov       idx1,    #line_longs      'perpare to move a whole row
:loop         
:mod          or       0-0,    0-0              'XOR source long to dest lon in COG memory
              add       :mod,   inc_sd           'increment the source and destination
              djnz      idx1,    #:loop          'loop for a whole row
or_line_ret   ret

''counts the number of "live" cells (set bits) in the frame buffer @data_ptr
''sets row_col equal to the number of "live" cells
''ala http://graphics.stanford.edu/~seander/bithacks.html#CountBitsSetParallel
{{unsigned int v; // count bits set in this (32-bit value)
unsigned int c; // store the total here
static const int S[] = {1, 2, 4, 8, 16}; // Magic Binary Numbers
static const int B[] = {0x55555555, 0x33333333, 0x0F0F0F0F, 0x00FF00FF, 0x0000FFFF};

c = v - ((v >> 1) & B[0]);
c = ((c >> S[1]) & B[1]) + (c & B[1]);
c = ((c >> S[2]) + c) & B[2];
c = ((c >> S[3]) + c) & B[3];
c = ((c >> S[4]) + c) & B[4];}}
census_asm
              mov       row_ptr, data_ptr       'start at the beginning of the array
              mov       row_idx, lcd_longs_asm  'setup to scan through whole array
              mov       row_col, #0             'reset output count
              'c = v - ((v >> 1) & B[0]); 
:loop         rdlong    t1,     row_ptr         't1 is 'v'
              add       row_ptr, #4             'advance to next long                
              mov       t2,     t1              '(v >> 1)
              shr       t2,     #1
              and       t2,     b_0             '( ~~ & B[0])
              sub       t1,     t2              't1 is now 'c'
              'c = ((c >> S[1]) & B[1]) + (c & B[1]);
              mov       t2,     t1
              shr       t2,     #2
              and       t2,     b_1
              and       t1,     b_1
              add       t1,     t2
              'c = ((c >> S[2]) + c) & B[2];
              mov       t2,     t1
              shr       t2,     #4
              add       t1,     t2
              and       t1,     b_2
              'c = ((c >> S[3]) + c) & B[3];
              mov       t2,     t1
              shr       t2,     #8
              add       t1,     t2
              and       t1,     b_3
              'c = ((c >> S[4]) + c) & B[4];
              mov       t2,     t1
              shr       t2,     #16
              add       t1,     t2
              and       t1,     b_4
              'done total 'c' is in 't1'
              add       row_col, t1             'accumulate the population from this long
              djnz      row_idx, #:loop         'loop again
              'write back total
              mov       t1,     par
              sub       t1,     #4              'address of row_col in hub
              wrlong    row_col, t1             'write out result  
census_asm_ret ret
'magic numbers
b_0     long  $5555_5555
b_1     long  $3333_3333
b_2     long  $0f0f_0f0f
b_3     long  $00ff_00ff
b_4     long  $0000_ffff

''writes numbers and letters to the frame buffer @data_ptr
''uses location and charicter code multiplexed into row_col
text_asm


text_asm_ret ret



'command interface variables
data_ptr long 0-0       'read from hub, pointer to LCD frame-buffer. 
row_col  long 0-0       'rw from hub, muxed row(s) and collum(s) bit index to start drawing at. (up to 4 parameters muxed into this)
command  long 0-0       '[set PAR here] a non-zero value indicated that a command has been issued.  re-set to zero by assembly when command is completed
                        'command == 1 : calculate the next generation of life on the frame-buffer @data_ptr
                                'assumes frame buffer is padded with space for bc_start and bc_end. 
                        'command == 2 : calculate number of ones in the frame buffer.  (result written to row_col)
                        'command == 3 : blit from font table.  three bytes packed in row_col == [0, font_code, start_row, start_col]
'command value variables
life_cmd   long 1
census_cmd long 2
text_cmd   long 3

'useful constants
inc_s   long  1         'add to an instruction to increment the source
inc_d   long  |< 9      'add to an instruction to increment the destination
inc_sd  long (|<9) | 1  'add to an instruction to increment the source AND destination.
                        

'constants needed by assembly code
msb_mask long |<31
'lcd_p   long  0-0       'pixels per line
'lcd_l   long  0-0       'lines per display
lcd_longs_asm long lcd_longs
lcd_bytes_asm long lcd_longs*4     'number of longs in LCD frame-buffer
'line_longs long 0-0     'number of longs in a line                           
all_one long -1         'a constant with every bit set.
        long -1
        long -1
        long -1                        

'general use variables
t1      res  1           
t2      res  1
t3      res  1
t4      res  1
idx1    res  1
idx2    res  1
t_row   res   line_longs


'variables use by life-solver
  nw    res  line_longs                      'direction longs used in array update calculations
  n     res  line_longs                       'nw n  ne
  ne    res  line_longs                      'w  ct e
  w     res  line_longs                       'sw s  se
  e     res  line_longs
  sw    res  line_longs
  s     res  line_longs
  se    res  line_longs
  ct    res  line_longs                      'space for center, at the end because it's not added to the parallel counter
  s_1   res  1                    'one component of a 3-bit by 32 parallel counter
  s_2   res  1
  s_3   res  1
  carry_1     res 1                 'carry variables for parallel counter
  carry_2     res 1
  carry_3     res 1     'aka s_4 (only calculated the last round)
  row_ptr     res 1
  row         res 1
  row_idx     res 1
  col_ptr     res 1
  col         res 1
  col_idx     res 1
  add_ptr     res 1
  add_idx     res 1
  

fit


con
  junk = |<9 | 1


        {'sort of rule 110 in 2D
              't_row := n
              mov       src_ptr, #n
              mov       dest_ptr, #t_row
              call      #mov_line
              't_row ^= ne
              mov       src_ptr, #ne
              call      #xor_line
              'w ^= all_one
              mov       src_ptr, #all_one
              mov       dest_ptr, #w
              call      #xor_line
              'n &= nw
              mov       src_ptr, #w
              mov       dest_ptr, #n
              call      #and_line
              'n &= ne
              mov       src_ptr, #ne
              call      #and_line
              'n |= t_row
              mov       src_ptr, #t_row
              call      #or_line              
              'life_array[row] := n
              mov       src_ptr, #n
              mov       dest_ptr, row_ptr
              call      #mov_line_to_hub                '}